#![no_std]
#![deny(missing_docs)]

//! Lock-free circular buffer

use core::fmt::Write;
use core::marker::PhantomData;
use core::mem::{self, MaybeUninit};
use core::ops::Range;
use core::ptr::NonNull;
use core::slice;
use core::sync::atomic::{AtomicUsize, Ordering};

/// Circular buffer. The `'a` lifetime represents the lifetime of the
/// underlying storage buffer, and `T` the type of stored data.
///
/// The buffer can have one writer and one reader at any time. No locks are
/// involved, and reader and writer cannot prevent each other from working
/// if there are enough room and data available.
///
/// The buffer works as follows: up to `capacity`-1 objects can be stored.
/// The ready-to-read area is `read_index`..`write_index` (modulo capacity)
/// The ready-to-write area is `write_index`..`read_index-1` (modulo capacity)
/// If `write_index == read_index` then there is nothing to read. If
/// `write_index + 1 == read_index` (modulo capacity) then there is no room
/// to write.
pub struct CircularBuffer<'a, T> {
    buffer: NonNull<T>,
    capacity: usize,
    read_index: AtomicUsize,
    write_index: AtomicUsize,
    _phantom_data: PhantomData<&'a T>,
}

// T must be Send for CircularBuffer to be Send, as it might
// already have data in it which will end up on another thread.
unsafe impl<'a, T: Send> Send for CircularBuffer<'a, T> {}

// T must be Send for CircularBuffer to be Sync, as other threads
// might use a reference on CircularBuffer to pop() a T and get
// it there.
unsafe impl<'a, T: Send> Sync for CircularBuffer<'a, T> {}

/// Error returned by read or write operations.
#[derive(Debug, Eq, PartialEq)]
pub enum Error {
    /// Trying to pretend that too much has been read.
    ReadLenError,
    /// Write would overflow.
    WouldOverflow,
}

impl<'a, T> CircularBuffer<'a, T> {
    /// Create circular buffer from existing storage.
    pub fn from_raw_parts(buffer: NonNull<T>, capacity: usize) -> Self {
        CircularBuffer {
            buffer,
            capacity,
            read_index: AtomicUsize::new(0),
            write_index: AtomicUsize::new(0),
            _phantom_data: PhantomData,
        }
    }

    /// Create circular buffer from existing storage.
    pub fn from_storage(buffer: &'a mut [MaybeUninit<T>]) -> Self {
        Self::from_raw_parts(unsafe { mem::transmute(buffer.as_mut_ptr()) }, buffer.len())
    }

    /// Buffer capacity.
    pub fn capacity(&self) -> usize {
        self.capacity
    }

    /// Return the indices at which data can be written, in order. The second
    /// item of the pair is present only when there are two disjoint ranges
    /// in which items can be written because it goes through the end of the buffer.
    fn write_indices(&self) -> (Range<usize>, Option<Range<usize>>) {
        todo!()
    }

    /// Number of entries that can be entered into the buffer without error.
    pub fn write_available(&self) -> usize {
        let (first, second) = self.write_indices();
        first.len() + second.map(|r| r.len()).unwrap_or_default()
    }

    /// Number of entries currently in the buffer.
    pub fn len(&self) -> usize {
        self.capacity - 1 - self.write_available()
    }

    /// Make all data in the circular buffer read.
    pub fn clear(&self) {
        while self.execute_some(|r| r.len(), None) > 0 {}
    }

    /// Check if the buffer is full
    pub fn is_full(&self) -> bool {
        self.write_available() == 0
    }

    /// Check if the buffer is empty
    pub fn is_empty(&self) -> bool {
        self.len() == 0
    }

    /// Write one element into the buffer.
    pub fn push(&self, data: T) -> Result<(), Error> {
        self.push_all([data])
    }

    /// Write data into the circular buffer. Up to the capacity minus one
    /// item can be written.
    /// Either everything is written or nothing, which is why we require
    /// an [`ExactSizeIterator`].
    pub fn push_all<I, II>(&self, data: I) -> Result<(), Error>
    where
        I: IntoIterator<IntoIter = II>,
        II: Iterator<Item = T>,
        II: ExactSizeIterator,
    {
        // Hint: use self.write_indices() in order to get where to write the data.
        todo!()
    }

    /// Read one element from the buffer if any.
    pub fn pop(&self) -> Option<T> {
        todo!()
    }

    /// Return a range with the indices of consecutive data to be read.
    fn read_indices(&self) -> Range<usize> {
        todo!()
    }

    /// Mark `size` consecutive entries as read.
    fn mark_read(&self, size: usize) -> Result<(), Error> {
        todo!()
    }

    /// Execute a function on some or all the consecutively available data.
    /// If T is not Sync, this function must not use the data but can query
    /// the slice (e.g., to get its length).
    fn execute_some<R, F>(&self, f: F, max: Option<usize>) -> R
    where
        F: FnOnce(&[T]) -> R,
    {
        let mut indices = self.read_indices();
        if let Some(max) = max {
            indices.end = indices.end.min(indices.start + max);
        }
        let data = unsafe {
            slice::from_raw_parts(
                self.buffer.as_ptr().add(indices.start) as *const T,
                indices.len(),
            )
        };
        let result = f(data);
        self.mark_read(data.len()).unwrap();
        result
    }
}

// Working on &T requires T to be Sync in addition. It is already Send if the
// CircularBuffer has been referenced from another thread so there is no need
// to add this bound again.
impl<'a, T: Sync> CircularBuffer<'a, T> {
    /// Execute a function on the consecutively available data then discard it.
    pub fn execute<R, F>(&self, f: F) -> R
    where
        F: FnOnce(&[T]) -> R,
    {
        self.execute_some(f, None)
    }

    /// Execute a function on the consecutively available data with a maximum size
    /// then discard it.
    pub fn execute_max<R, F>(&self, max_len: usize, f: F) -> R
    where
        F: FnOnce(&[T]) -> R,
    {
        self.execute_some(f, Some(max_len))
    }
}

impl<'a, T: Clone> CircularBuffer<'a, T> {
    /// Write data into the circular buffer. Up to the capacity minus one
    /// item can be written. Either everything is written or nothing.
    pub fn write(&self, data: &[T]) -> Result<(), Error> {
        self.push_all(data.iter().cloned())
    }
}

impl<'a, T> Drop for CircularBuffer<'a, T> {
    fn drop(&mut self) {
        todo!()
    }
}

// Iterator is implemented on &CircularBuffer rather
// than CircularBuffer because combinators consume their
// arguments and CircularBuffer is not clonable.
impl<'a, T> Iterator for &CircularBuffer<'a, T> {
    type Item = T;

    fn next(&mut self) -> Option<Self::Item> {
        self.pop()
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        (self.len(), None)
    }
}

// Write is implemented on &CircularBuffer instead of CircularBuffer
// to avoid needing a mutable reference to it.
impl Write for &CircularBuffer<'_, u8> {
    fn write_str(&mut self, s: &str) -> core::fmt::Result {
        self.push_all(s.bytes()).map_err(|_| core::fmt::Error)
    }
}
